import PyIPSDK
import PyIPSDK.IPSDKIPLIntensityTransform as itrans
import PyIPSDK.IPSDKIPLArithmetic as arithm
import PyIPSDK.IPSDKIPLGlobalMeasure as glbmsr
import PyIPSDK.IPSDKIPLUtility as util
import PyIPSDK.IPSDKIPLBinarization as bin
import PyIPSDK.IPSDKIPLLogical as logic


def processFiltering(inImg_, nbIter_, sigmaX_, sigmaY_):

    def moduleSquareComplexImage(image):

        reSquare = arithm.squareImg(image[0])
        imSquare = arithm.squareImg(image[1])
        moduleImage = arithm.addImgImg(reSquare, imSquare)

        return moduleImage

    def conjugateComplexeImage(image):

        Re = util.copyImg(image[0])
        Im = arithm.multiplyScalarImg(image[1],-1)

        return [Re,Im]

    def multiplyComplexImages(image1,image2):

        image_Re_1 = arithm.multiplyImgImg(image1[0], image2[0])
        if image1[1] != None and image2[1] != None:
            image_Re_2 = arithm.multiplyImgImg(image1[1], image2[1])
            outImage_Re = arithm.subtractImgImg(image_Re_1, image_Re_2)
        else:
            outImage_Re = util.copyImg(image_Re_1)
        if image1[1] == None:
            outImage_Im = arithm.multiplyImgImg(image1[0], image2[1])
        elif image2[1] == None:
            outImage_Im = arithm.multiplyImgImg(image1[1], image2[0])
        else:
            image_Im_1 = arithm.multiplyImgImg(image1[1], image2[0])
            image_Im_2 = arithm.multiplyImgImg(image1[0], image2[1])
            outImage_Im = arithm.addImgImg(image_Im_1, image_Im_2)

        return [outImage_Re,outImage_Im]

    def addComplexImages(image1,image2):

        outImage_Re = arithm.addImgImg(image1[0],image2[0])
        outImage_Im = arithm.addImgImg(image1[1],image2[1])

        return [outImage_Re,outImage_Im]

    def getFayImage(y1Img,y2Img,img1_F,img2_F,filterImg_Module,dftConfig):

        y1Img_F = itrans.forwardDiscreteFourierTransform2dImg(y1Img, dftConfig)
        y2Img_F = itrans.forwardDiscreteFourierTransform2dImg(y2Img, dftConfig)

        img1C_F = conjugateComplexeImage(img1_F)
        img1C_y1Img_F = multiplyComplexImages(img1C_F, y1Img_F)

        img2C_F = conjugateComplexeImage(img2_F)
        img2C_y2Img_F = multiplyComplexImages(img2C_F, y2Img_F)

        sumImage = addComplexImages(img1C_y1Img_F, img2C_y2Img_F)

        fayImg_F = multiplyComplexImages(sumImage, [filterImg_Module, None])

        fayImg_F = multiplyComplexImages(fayImg_F, [filterImg_Module, None])

        return fayImg_F

    # gabor image function generation
    def generateGaborImg(inImg, sigmaX, sigmaY):
        width = inImg.getSizeX()
        height = inImg.getSizeY()
        cx = int(width / 2) + 1
        cy = int(height / 2) + 1
        sigmaX2 = sigmaX * sigmaX
        sigmaY2 = sigmaY * sigmaY
        xStr = "(" + str(cx) + "-x)"
        x2Str = xStr + "*" + xStr
        yStr = "(" + str(cy) + "-y)"
        y2Str = yStr + "*" + yStr
        formulaStr = "exp(-0.5 * (" + x2Str + "/" + str(sigmaX2) + " + " + y2Str + "/" + str(sigmaY2) + "))"
        gaborImg = arithm.formula2dImg(formulaStr, InOptImg1=inImg)

        return gaborImg

    # filtering function

    filterImg = generateGaborImg(inImg_, sigmaX_, sigmaY_)

    alpha = 1

    dftConfig = PyIPSDK.createDFTConfig(PyIPSDK.eDFTQuadrantsPolicy.eDFTQP_Native, PyIPSDK.eDFTCoordinates.eDFTC_Cartesian, PyIPSDK.eDFTScalePolicy.eDFTSP_Default)
    nbRows = inImg_.getSizeY()
    nbCols = inImg_.getSizeX()

    img1 = PyIPSDK.createImage(PyIPSDK.eImageBufferType.eIBT_Real32, nbCols, nbRows)
    util.eraseImg(img1,0)
    img1.array[0, 0] = 1
    img1.array[nbRows - 1, 0] = -1

    img2 = PyIPSDK.createImage(PyIPSDK.eImageBufferType.eIBT_Real32, nbCols, nbRows)
    util.eraseImg(img2,0)
    img2.array[0, 0] = 1
    img2.array[0, nbCols - 1] = -1

    img1_F = itrans.forwardDiscreteFourierTransform2dImg(img1, dftConfig)
    img1 = None
    img2_F = itrans.forwardDiscreteFourierTransform2dImg(img2, dftConfig)
    img2 = None
    inImg_F = itrans.forwardDiscreteFourierTransform2dImg(inImg_, dftConfig)
    filterImg_F = itrans.forwardDiscreteFourierTransform2dImg(filterImg, dftConfig)
    filterImg_Module_Square = moduleSquareComplexImage(filterImg_F)
    filterImg_F = None

    d1_u0_Img_F = multiplyComplexImages(img1_F, inImg_F)
    d1_u0_Img = itrans.backwardDiscreteFourierTransform2dImg(d1_u0_Img_F[0], d1_u0_Img_F[1], dftConfig)
    d1_u0_Img_F = None
    d2_u0_Img_F = multiplyComplexImages(img2_F, inImg_F)
    inImg_F = None
    d2_u0_Img = itrans.backwardDiscreteFourierTransform2dImg(d2_u0_Img_F[0], d2_u0_Img_F[1], dftConfig)
    d2_u0_Img_F = None

    img1_Module_Square = moduleSquareComplexImage(img1_F)
    img2_Module_Square = moduleSquareComplexImage(img2_F)
    img1_img2_Module_Square = arithm.addImgImg(img1_Module_Square, img2_Module_Square)
    img1_Module_Square = None
    img2_Module_Square = None
    filter_img1_img2_Module_Square = arithm.multiplyImgImg(img1_img2_Module_Square, filterImg_Module_Square)
    img1_img2_Module_Square = None
    maxValue = glbmsr.statsMsr2d(filter_img1_img2_Module_Square).max
    filter_img1_img2_Module_Square = None
    L = 2 / alpha * maxValue
    #print("L=" + str(L))

    y1Img = PyIPSDK.createImage(PyIPSDK.eImageBufferType.eIBT_Real32, nbCols, nbRows)
    util.eraseImg(y1Img,0)
    y2Img = PyIPSDK.createImage(PyIPSDK.eImageBufferType.eIBT_Real32, nbCols, nbRows)
    util.eraseImg(y2Img, 0)
    q1Img = PyIPSDK.createImage(PyIPSDK.eImageBufferType.eIBT_Real32, nbCols, nbRows)
    util.eraseImg(q1Img, 0)
    q2Img = PyIPSDK.createImage(PyIPSDK.eImageBufferType.eIBT_Real32, nbCols, nbRows)
    util.eraseImg(q2Img, 0)

    for iter in range(0, nbIter_):

        #print("Iter=" + str(iter))

        y1Img_F = itrans.forwardDiscreteFourierTransform2dImg(y1Img, dftConfig)
        y2Img_F = itrans.forwardDiscreteFourierTransform2dImg(y2Img, dftConfig)

        img1C_F = conjugateComplexeImage(img1_F)
        img1C_y1Img_F = multiplyComplexImages(img1C_F, y1Img_F)
        y1Img_F = None
        img1C_F = None

        img2C_F = conjugateComplexeImage(img2_F)
        img2C_y2Img_F = multiplyComplexImages(img2C_F, y2Img_F)
        y2Img_F = None
        img2C_F = None

        sumImage = addComplexImages(img1C_y1Img_F, img2C_y2Img_F)
        img1C_y1Img_F = None
        img2C_y2Img_F = None

        fayImg_F = multiplyComplexImages(sumImage, [filterImg_Module_Square, None])
        fayImg_F[0] = arithm.multiplyScalarImg(fayImg_F[0],1/2)
        fayImg_F[1] = arithm.multiplyScalarImg(fayImg_F[1],1/2)
        sumImage = None

        q1p_re = util.copyImg(q1Img)
        q2p_re = util.copyImg(q2Img)

        fayImg_Img1_F = multiplyComplexImages(fayImg_F,img1_F)
        fayImg_Img1 = itrans.backwardDiscreteFourierTransform2dImg(fayImg_Img1_F[0], fayImg_Img1_F[1], dftConfig)
        fayImg_Img1_F = None
        q1Img = arithm.subtractImgImg(d1_u0_Img, fayImg_Img1)
        fayImg_Img1 = None

        fayImg_Img2_F = multiplyComplexImages(fayImg_F,img2_F)
        fayImg_Img2 = itrans.backwardDiscreteFourierTransform2dImg(fayImg_Img2_F[0], fayImg_Img2_F[1], dftConfig)
        fayImg_Img2_F = None
        q2Img = arithm.subtractImgImg(d2_u0_Img, fayImg_Img2)
        fayImg_Img2 = None

        q1Img = arithm.multiplyScalarImg(q1Img, 1 / L)
        q1Img = arithm.addImgImg(y1Img, q1Img)
        q2Img = arithm.multiplyScalarImg(q2Img, 1 / L)
        q2Img = arithm.addImgImg(y2Img, q2Img)
        normq1q2 = arithm.l2Norm2Img(q1Img, q2Img)

        maskImage = bin.thresholdImg(normq1q2, 1, PyIPSDK.getMaxValue(PyIPSDK.eImageBufferType.eIBT_Real32))
        q1ImgNormed = arithm.divideImgImg(q1Img, normq1q2)
        q1Img = logic.maskImgImg(q1ImgNormed, q1Img, maskImage)
        q1ImgNormed = None
        q2ImgNormed = arithm.divideImgImg(q2Img, normq1q2)
        normq1q2 = None
        q2Img = logic.maskImgImg(q2ImgNormed, q2Img, maskImage)
        q2ImgNormed = None
        maskImage = None

        factor = iter / (iter + 3.0)
        y1Img = arithm.subtractImgImg(q1Img, q1p_re)
        q1p_re = None
        y1Img = arithm.multiplyScalarImg(y1Img, factor)
        y1Img = arithm.addImgImg(y1Img, q1Img)
        y2Img = arithm.subtractImgImg(q2Img, q2p_re)
        q2p_re = None
        y2Img = arithm.multiplyScalarImg(y2Img, factor)
        y2Img = arithm.addImgImg(y2Img, q2Img)

    fayImg = itrans.backwardDiscreteFourierTransform2dImg(fayImg_F[0], fayImg_F[1], dftConfig)
    fayImg_F = None
    noiseImg = arithm.multiplyScalarImg(fayImg, 1 / alpha)
    fayImg = None
    filteredImg = arithm.subtractImgImg(inImg_, noiseImg)

    noiseImg = util.convertImg(noiseImg,inImg_.getBufferType())
    filteredImg = util.convertImg(filteredImg,inImg_.getBufferType())

    return filteredImg, noiseImg

# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------

# parameters
# nbIter = 12
# sigmaX = 1
# sigmaY = 30

# import glob
# import os
#
# inputFilename = "F:\Total\stack_for_registration"
# outputFilename = "F:\Total\stack_corrected"
#
# if os.path.exists(outputFilename) == False:
#     os.makedirs(outputFilename)
#
# filterImg = None
# for imageFilename in glob.glob(inputFilename + "/*.tif"):
#     inImg = PyIPSDK.loadTiffImageFile(imageFilename)
#     if filterImg is None:
#         filterImg = generateGaborImg(inImg, sigmaX, sigmaY)
#     filteredImg, noiseImg = processFiltering(inImg, nbIter, filterImg)
#
#     PyIPSDK.saveTiffImageFile(outputFilename + "/" + os.path.basename(imageFilename), filteredImg)
#
# inImg = PyIPSDK.loadTiffImageFile("F:/vsnr.tif")
# outImg, noiseImg = processFiltering(inImg,nbIter,sigmaX,sigmaY)
#
# import PyIPSDK.IPSDKUI as ui
# app = ui.getQApplication()
# ui.displayImg(inImg,"inImg")
# ui.displayImg(outImg,"Vsnr")
# app.exec_()




